!> \~japanese
!! \brief 数値計算に伴う誤差を評価するためのサブルーチンを提供するモジュール.
!!
!! \par Field_IcGrid2D_error_norm_check モジュールについて
!!
!! \~english
!! \brief This module provides some subroutines to evaluate the error for numerical methods.
!!
!! \par About Field_IcGrid2D_error_norm_check module
!!
!! \~
!!
!! <br><br>
!! \b Copyright (C) GFD Dennou Club, 2011-2012. All rights reserved. <br>
!! \b license ?? <br>
!! \author Yuta Kawai
!!
module Field_IcGrid2D_error_norm_check

  ! モジュール引用 ; Use statements
  !

  ! 種類型パラメータ
  ! Kind type parameter
  !
  use dc_types, only: DP


  ! 正二十面体格子データの管理
  ! Manager of icosahedral grid data
  !
  use IcGrid2D_FVM_Manager, only: &
    & IcGrid2D_FVM, &
    & get_EffSize_Min, get_EffSize_Max

  ! 正二十面体格子上の物理場データの管理
  ! Manager of the physical field data on icosahedral grid
  !
  use Field_IcGrid2D_Manager, only: &
    & Field_IcGrid2D, &
    & get_icgrid
  
  ! 全矩形領域における厳密解に対する数値解の誤差チェック
  !
  !
  use rcregions_norm_check, only: &
       & calc_norm1, calc_norm2, calc_norm3

  ! 宣言文 ; Declaration statements
  !
  implicit none
  private

  ! 公開手続き
  ! Public procedures
  !
  public :: numerical_error_norm1, numerical_error_norm2, numerical_error_norminfinity

contains

!
!> \~japanese
!! \brief 厳密解から数値解に対する誤差ノルム 1 を求める.
!!
!! @param[in] true_field ある場の厳密解を管理する構造型 Field_IcGrid2D の変数.
!! @param[in] num_field ある場の数値解を管理する構造型 Field_IcGrid2D の変数.
!! @return normVal 誤差ノルム 1.
!!
!! \~english
!! \brief Calculates the error norm 1 for the numerical solution using exact solution.
!!
!! @param[in] true_field The variable of derived type IcGrid Field which manages the exaact solution of a given field.
!! @param[in] num_field The variable of derived type IcGrid Field which manages the numerical solution of a given field.
!! @return normVal Error norm 1.
!!
function numerical_error_norm1( &
  & true_field, num_field &  ! (in)
  & ) result(normVal)

  ! 宣言文 ; Declaration statements
  !
  type(Field_IcGrid2D), intent(in) :: true_field
  type(Field_IcGrid2D), intent(in) :: num_field
  real(DP) normVal
  
  ! 作業変数
  ! Work variables
  !
  integer :: EMin, EMax
  type(IcGrid2D_FVM), pointer :: icgrid

  ! 実行文 ; Executable statements
  !
  icgrid => get_icgrid(true_field)
  EMin = get_EffSize_Min(icgrid)
  EMax = get_EffSize_Max(icgrid)

  normVal = calc_norm1( &
       & true_field%val(:,EMin:EMax, EMin:EMax, :), num_field%val(:,EMin:EMax, EMin:EMax, :) &
       & )

end function numerical_error_norm1

!
!> \~japanese
!! \brief 厳密解から数値解に対する誤差ノルム 2 を求める.
!!
!! @param[in] true_field ある場の厳密解を管理する構造型 Field_IcGrid2D の変数.
!! @param[in] num_field ある場の数値解を管理する構造型 Field_IcGrid2D の変数.
!! @return normVal 誤差ノルム 2.
!!
!! \~english
!! \brief Calculates the error norm 2 for the numerical solution using exact solution.
!!
!! @param[in] true_field The variable of derived type IcGrid Field which manages the exaact solution of a given field.
!! @param[in] num_field The variable of derived type IcGrid Field which manages the numerical solution of a given field.
!! @return normVal Error norm 2.
!!
function numerical_error_norm2( &
  & true_field, num_field &  ! (in)
  & ) result(normVal)

  ! 宣言文 ; Declaration statements
  !
  type(Field_IcGrid2D), intent(in) :: true_field
  type(Field_IcGrid2D), intent(in) :: num_field
  real(DP) normVal
  
  ! 作業変数
  ! Work variables
  !
  integer :: EMin, EMax
  type(IcGrid2D_FVM), pointer :: icgrid

  ! 実行文 ; Executable statements
  !
  icgrid => get_icgrid(true_field)
  EMin = get_EffSize_Min(icgrid)
  EMax = get_EffSize_Max(icgrid)

  normVal = calc_norm2( &
       & true_field%val(:,EMin:EMax, EMin:EMax, :), num_field%val(:,EMin:EMax, EMin:EMax, :) &
       & )

end function numerical_error_norm2

!
!> \~japanese
!! \brief 厳密解から数値解に対する誤差ノルム 3 を求める.
!!
!! @param[in] true_field ある場の厳密解を管理する構造型 Field_IcGrid2D の変数.
!! @param[in] num_field ある場の数値解を管理する構造型 Field_IcGrid2D の変数.
!! @param[in] maxmin_info_flag 厳密解と数値解の局所的なずれが最も大きい位置および最も小さい位置の情報を出力するかのフラグ.
!! デフォルトは false である.
!! @return normVal 誤差ノルム 3.
!!
!! \~english
!! \brief Calculates the error norm 3 for the numerical solution using exact solution.
!!
!! @param[in] true_field The variable of derived type IcGrid Field which manages the exaact solution of a given field.
!! @param[in] num_field The variable of derived type IcGrid Field which manages the numerical solution of a given field.
!! @param[in] maxmin_info_flag The flag indicating whether one needs the information about the location
!! where the local difference with exact and numerical solution is maximum or minimum. The default flag value is false.
!! @return normVal Error norm 3.
!!
function numerical_error_norminfinity( &
  & true_field, num_field, maxmin_info_flag &  ! (in)
  & ) result(normVal)

  ! 宣言文 ; Declaration statements
  !
  type(Field_IcGrid2D), intent(in) :: true_field
  type(Field_IcGrid2D), intent(in) :: num_field
  logical, intent(in), optional :: maxmin_info_flag
  real(DP) normVal
  
  ! 作業変数
  ! Work variables
  !
  integer :: EMin, EMax
  type(IcGrid2D_FVM), pointer :: icgrid
  logical :: info_flag

  ! 実行文 ; Declaration statements
  !
  icgrid => get_icgrid(true_field)
  EMin = get_EffSize_Min(icgrid)
  EMax = get_EffSize_Max(icgrid)

  if ( present(maxmin_info_flag) .and. maxmin_info_flag ) then
    info_flag = .true.
  else
    info_flag = .false.
  end if

  normVal = calc_norm3( &
       & true_field%val(:,EMin:EMax, EMin:EMax, :), num_field%val(:,EMin:EMax, EMin:EMax, :), &
       & output_info_flag=info_flag )

end function numerical_error_norminfinity

end module Field_IcGrid2D_error_norm_check
